import os

import lightgbm
import numpy as np
import pandas as pd
import pytest

import lleaves


@pytest.mark.skipif(
    os.environ.get("CI", "false") == "true",
    reason="We don't want to download the datasets on CI",
)
def test_benchmark_datsets_correct_output():
    from benchmarks.benchmark import NYC_used_columns
    from benchmarks.train_NYC_model import feature_enginering

    model_file_NYC = "tests/models/NYC_taxi/model.txt"
    model_file_airline = "tests/models/airline/model.txt"

    df = pd.read_parquet(
        "benchmarks/data/yellow_tripdata_2016-01.parquet", columns=NYC_used_columns
    )
    NYC_X = feature_enginering().fit_transform(df).astype(np.float32)

    df = pd.read_csv("benchmarks/data/airline_data_factorized.csv")
    airline_X = df.to_numpy(np.float32)

    for model_file, data in [(model_file_NYC, NYC_X), (model_file_airline, airline_X)]:
        lgb = lightgbm.Booster(model_file=model_file)
        llvm = lleaves.Model(model_file=model_file)
        llvm.compile()
        np.testing.assert_almost_equal(
            lgb.predict(data, n_jobs=4), llvm.predict(data, n_jobs=4), decimal=13
        )



from ctypes import POINTER, c_double

import numpy as np
from lightgbm import Booster

from lleaves import Model


def test_parallel_iteration():
    llvm_model = Model(model_file="tests/models/NYC_taxi/model.txt")
    lgbm_model = Booster(model_file="tests/models/NYC_taxi/model.txt")
    llvm_model.compile()

    data = np.array(4 * [5 * [1.0]], dtype=np.float64)
    data_flat = np.array(data.reshape(data.size), dtype=np.float64)
    np.testing.assert_almost_equal(
        llvm_model.predict(data, n_jobs=4), lgbm_model.predict(data), decimal=14
    )

    ptr_data = data_flat.ctypes.data_as(POINTER(c_double))
    preds = np.zeros(4, dtype=np.float64)
    ptr_preds = preds.ctypes.data_as(POINTER(c_double))

    llvm_model._c_entry_func(ptr_data, ptr_preds, 2, 4)
    preds_l = list(preds)
    assert preds_l[0] == 0.0 and preds_l[1] == 0.0
    assert preds_l[2] != 0.0 and preds_l[3] != 0.0
    llvm_model._c_entry_func(ptr_data, ptr_preds, 0, 2)
    preds_l = list(preds)
    assert preds_l[0] != 0.0 and preds_l[1] != 0.0



from lleaves.compiler.utils import DecisionType


class Forest:
    def __init__(
        self,
        trees: list,
        features: list,
        objective_func: str,
        objective_func_config: str,
    ):
        self.trees = trees
        self.n_args = len(features)
        self.features = features
        self.objective_func = objective_func
        self.objective_func_config = objective_func_config


class Tree:
    def __init__(self, idx, root_node, features):
        self.idx = idx
        self.root_node = root_node
        self.features = features

    def __str__(self):
        return f"tree_{self.idx}"


class Node:
    @property
    def is_leaf(self):
        return isinstance(self, LeafNode)


class DecisionNode(Node):
    # the threshold in bit-representation if this node is categorical
    cat_threshold = None

    # child nodes
    left = None
    right = None

    def __init__(
        self,
        idx: int,
        split_feature: int,
        threshold: int,
        decision_type_id: int,
        left_idx: int,
        right_idx: int,
    ):
        self.idx = idx
        self.split_feature = split_feature
        self.threshold = threshold
        self.decision_type = DecisionType(decision_type_id)
        self.right_idx = right_idx
        self.left_idx = left_idx

    def add_children(self, left, right):
        self.left = left
        self.right = right

    def finalize_categorical(self, cat_threshold):
        self.cat_threshold = cat_threshold
        self.threshold = int(self.threshold)

    def validate(self):
        if self.decision_type.is_categorical:
            assert self.cat_threshold is not None
        else:
            assert self.threshold

    def __str__(self):
        return f"node_{self.idx}"


class LeafNode(Node):
    def __init__(self, idx, value):
        self.idx = idx
        self.value = value

    def __str__(self):
        return f"leaf_{self.idx}"



import lightgbm as lgb
import numpy as np
import pytest

import lleaves


def test_interface(tmp_path):
    lgbm = lgb.Booster(model_file="tests/models/tiniest_single_tree/model.txt")
    llvm = lleaves.Model("tests/models/tiniest_single_tree/model.txt")
    llvm.compile()

    for arr in [
        np.array([1.0, 1.0, 1.0]),
        [1.0, 1.0, 1.0],
    ]:
        with pytest.raises(ValueError) as err1:
            llvm.predict(arr)
        with pytest.raises(ValueError) as err2:
            lgbm.predict(arr)
        assert "dimension" in err1.value.args[0]
        assert "dimension" in err2.value.args[0]

    with pytest.raises(ValueError):
        wrong_shape_2D = (np.array(3 * [(llvm.num_feature() + 1) * [1.0]]),)
        llvm.predict(wrong_shape_2D)


@pytest.mark.parametrize(
    "model_file, n_args",
    [
        ("tests/models/pure_categorical/model.txt", 3),
        ("tests/models/tiniest_single_tree/model.txt", 3),
    ],
)
def test_input_dtypes(model_file, n_args):
    lgbm = lgb.Booster(model_file=model_file)
    llvm = lleaves.Model(model_file)
    llvm.compile()

    arr = np.array([[1.0, 1.0, 1.0]], dtype=np.float32)
    assert llvm.predict(arr) == lgbm.predict(arr)
    arr = np.array([[1.0, 1.0, 1.0]], dtype=np.float64)
    assert llvm.predict(arr) == lgbm.predict(arr)
    arr = np.array([[0, 0, 0]], dtype=np.int32)
    assert llvm.predict(arr) == lgbm.predict(arr)
    arr = [[0, 0, 0]]
    assert llvm.predict(arr) == lgbm.predict(arr)


def test_cache_model(tmp_path):
    cachefp = tmp_path / "model.bin"
    pure_cat_llvm = lleaves.Model("tests/models/pure_categorical/model.txt")
    assert not cachefp.exists()
    pure_cat_llvm.compile(cache=cachefp)
    assert cachefp.exists()
    # we compiled model.txt to IR to asm
    res = pure_cat_llvm.predict([3 * [0.0], 3 * [1.0], 3 * [-1.0]])

    cached_model = lleaves.Model("tests/models/tiniest_single_tree/model.txt")
    cached_model.compile(cache=cachefp)

    tiniest_llvm = lleaves.Model("tests/models/tiniest_single_tree/model.txt")
    tiniest_llvm.compile()

    # the cache was loaded (which was different from the model.txt passed)
    np.testing.assert_equal(
        cached_model.predict([3 * [0.0], 3 * [1.0], 3 * [-1.0]]),
        pure_cat_llvm.predict([3 * [0.0], 3 * [1.0], 3 * [-1.0]]),
    )
    # sanity test
    np.testing.assert_equal(
        pure_cat_llvm.predict([3 * [0.0], 3 * [1.0], 3 * [-1.0]]), res
    )



import lightgbm as lgb
import numpy as np
import pytest

from lleaves import Model


@pytest.fixture(
    params=[
        "binary sigmoid:10",
        "binary sigmoid:3",
        "binary sigmoid:0.1",
        "cross_entropy",
        "xentropy",
        "xentlambda",
        "cross_entropy_lambda",
        "poisson",
        "gamma",
        "tweedie",
        "regression sqrt",
        "regression",
    ]
)
def modified_model_txt(request, tmp_path):
    model_filep = tmp_path / request.param.replace(" ", "_")
    with open("tests/models/leaf_scan/model.txt") as modelfile, open(
        model_filep, "w"
    ) as tmpfile:
        for line in modelfile:
            if line.startswith("objective="):
                tmpfile.write("objective=" + request.param + "\n")
            else:
                tmpfile.write(line)
    return str(model_filep)


def test_all_obj_funcs(modified_model_txt):
    data = np.expand_dims(np.arange(-5, 5, 0.20), axis=1)
    llvm_model = Model(model_file=modified_model_txt)
    llvm_model.compile()
    lgbm_model = lgb.Booster(model_file=modified_model_txt)
    np.testing.assert_almost_equal(lgbm_model.predict(data), llvm_model.predict(data))


@pytest.mark.parametrize("objective", ["regression", "binary"])
def test_basic(tmp_path, objective):
    X = np.expand_dims(np.array([1, 2, 3, 1, 2, 3, 1, 2, 3]), axis=1)
    y = np.array([1, 0, 0, 1, 0, 0, 1, 0, 0])
    train_data = lgb.Dataset(X, label=y, categorical_feature=[0])
    params = {"objective": objective}
    if objective == "multiclass":
        params["num_class"] = 2
    bst = lgb.train(params, train_data, 1, categorical_feature=[0])

    reg_model_f = str(tmp_path / f"{objective}.txt")
    bst.save_model(reg_model_f)
    llvm_model = Model(model_file=reg_model_f)
    llvm_model.compile()
    np.testing.assert_almost_equal(bst.predict(X), llvm_model.predict(X))



# code stolen from https://github.com/xhochy/nyc-taxi-fare-prediction-deployment-example
# download the data from https://www1.nyc.gov/site/tlc/about/tlc-trip-record-data.page
# use the pandas code snippet from here
# https://github.com/xhochy/nyc-taxi-fare-prediction-deployment-example/blob/main/training/Train.ipynb
# to convert the CSV to parquet

import lightgbm
import numpy as np
import pandas as pd
from sklearn.compose import make_column_transformer
from sklearn.preprocessing import FunctionTransformer


def haversine_distance(lat1, lng1, lat2, lng2):
    lat1, lng1, lat2, lng2 = (np.radians(x) for x in (lat1, lng1, lat2, lng2))
    d = (
        np.sin(lat2 / 2 - lat1 / 2) ** 2
        + np.cos(lat1) * np.cos(lat2) * np.sin(lng2 / 2 - lng1 / 2) ** 2
    )
    return 2 * 6371 * np.arcsin(np.sqrt(d))  # 6,371 km is the earth radius


def haversine_distance_from_df(df):
    return pd.DataFrame(
        {
            "haversine_distance": haversine_distance(
                df["pickup_latitude"],
                df["pickup_longitude"],
                df["dropoff_latitude"],
                df["dropoff_longitude"],
            )
        }
    )


def split_pickup_datetime(df):
    return pd.DataFrame(
        {
            "pickup_dayofweek": df["tpep_pickup_datetime"].dt.dayofweek,
            "pickup_hour": df["tpep_pickup_datetime"].dt.hour,
            "pickup_minute": df["tpep_pickup_datetime"].dt.minute,
        }
    )


def feature_enginering():
    return make_column_transformer(
        (FunctionTransformer(), ["passenger_count"]),
        (
            FunctionTransformer(func=split_pickup_datetime),
            ["tpep_pickup_datetime"],
        ),
        (
            FunctionTransformer(
                func=haversine_distance_from_df,
            ),
            [
                "pickup_latitude",
                "pickup_longitude",
                "dropoff_latitude",
                "dropoff_longitude",
            ],
        ),
    )


if __name__ == "__main__":
    used_columns = [
        "fare_amount",
        "pickup_latitude",
        "pickup_longitude",
        "dropoff_latitude",
        "dropoff_longitude",
        "tpep_pickup_datetime",
        "passenger_count",
    ]
    df = pd.read_parquet("data/yellow_tripdata_2016-01.parquet", columns=used_columns)
    y = df.pop("fare_amount")

    # feature_enginering is the example pipeline without the Regressor
    features_as_array = feature_enginering().fit_transform(df)
    train_data = lightgbm.Dataset(features_as_array, label=y)
    model = lightgbm.train({"objective": "regression_l1"}, train_data)
    model.save_model("../tests/models/model.txt")



import os
import time

import lightgbm
import numpy as np
import onnxmltools
import onnxruntime as rt
import pandas as pd
import treelite
import treelite_runtime
from onnxconverter_common import FloatTensorType

from benchmarks.train_NYC_model import feature_enginering
from lleaves import Model


class BenchmarkModel:
    model = None
    name = None

    def __init__(self, lgbm_model_file):
        self.model_file = lgbm_model_file

    def setup(self, data, n_threads):
        start = time.perf_counter()
        self._setup(data, n_threads)
        print(f"{self.name} setup: {round(time.perf_counter() - start, 2)}")

    def _setup(self, data, n_threads):
        raise NotImplementedError()

    def predict(self, data, index, batchsize, n_threads):
        return self.model.predict(data[index : index + batchsize], n_jobs=n_threads)

    def __str__(self):
        return self.name


class LGBMModel(BenchmarkModel):
    name = "LightGBM Booster"

    def _setup(self, data, n_threads):
        self.model = lightgbm.Booster(model_file=self.model_file)


class LLVMModel(BenchmarkModel):
    name = "lleaves"

    def _setup(self, data, n_threads):
        self.model = Model(model_file=self.model_file)
        self.model.compile()


class TreeliteModel(BenchmarkModel):
    name = "Treelite"

    def _setup(self, data, n_threads):
        # disable thread pinning, which modifies (and never resets!) process-global pthreads state
        os.environ["TREELITE_BIND_THREADS"] = "0"
        treelite_model = treelite.Model.load(self.model_file, model_format="lightgbm")
        treelite_model.export_lib(
            toolchain="gcc",
            libpath="/tmp/treelite_model.so",
            params={"parallel_comp": 4},
            verbose=False,
        )
        self.model = treelite_runtime.Predictor(
            "/tmp/treelite_model.so",
            nthread=n_threads,
        )

    def predict(self, data, index, batchsize, n_threads):
        return self.model.predict(
            treelite_runtime.DMatrix(data[index : index + batchsize])
        )


class ONNXModel(BenchmarkModel):
    name = "ONNX Runtime"

    def _setup(self, data, n_threads):
        lgbm_model = lightgbm.Booster(model_file=self.model_file)
        onnx_model = onnxmltools.convert_lightgbm(
            lgbm_model,
            initial_types=[
                (
                    "float_input",
                    FloatTensorType([None, lgbm_model.num_feature()]),
                )
            ],
            target_opset=8,
        )
        onnxmltools.utils.save_model(onnx_model, "/tmp/model.onnx")
        options = rt.SessionOptions()
        options.inter_op_num_threads = n_threads
        options.intra_op_num_threads = n_threads
        self.model = rt.InferenceSession("/tmp/model.onnx", sess_options=options)
        self.input_name = self.model.get_inputs()[0].name
        self.label_name = self.model.get_outputs()[0].name

    def predict(self, data, index, batchsize, n_threads):
        return self.model.run(
            [self.label_name], {self.input_name: data[index : index + batchsize]}
        )


NYC_used_columns = [
    "fare_amount",
    "pickup_latitude",
    "pickup_longitude",
    "dropoff_latitude",
    "dropoff_longitude",
    "tpep_pickup_datetime",
    "passenger_count",
]


def run_benchmark(
    model_files, np_data, model_classes, threadcount, batchsizes, n_samples=1000
):
    for model_file, data in zip(model_files, np_data):
        model_name = model_file.split("/")[-2]
        print(model_file, f"\n---- {str.upper(model_name)} --- \n")
        for n_threads in threadcount:
            for model_class in model_classes:
                model = model_class(model_file)
                results = {"time (μs)": [], "batchsize": []}
                model.setup(data, n_threads)
                for batchsize in batchsizes:
                    times = []
                    for _ in range(n_samples):
                        start = time.perf_counter_ns()
                        model.predict(data, 0, batchsize, n_threads)
                        # calc per-batch times, in μs
                        times.append((time.perf_counter_ns() - start) / 1000)
                    results["time (μs)"] += times
                    results["batchsize"] += len(times) * [batchsize]
                    print(
                        f"{model} (Batchsize {batchsize}, nthread {n_threads}): {round(min(times), 2)}μs"
                    )
        print()


if __name__ == "__main__":
    df = pd.read_parquet(
        "data/yellow_tripdata_2016-01.parquet", columns=NYC_used_columns
    )
    NYC_X = feature_enginering().fit_transform(df).astype(np.float32)

    df = pd.read_csv("data/airline_data_factorized.csv")
    airline_X = df.to_numpy(np.float32)

    df = pd.read_parquet("data/mtpl2.parquet")
    mtpl2_X = df.to_numpy(np.float32)

    model_file_NYC = "../tests/models/NYC_taxi/model.txt"
    model_file_airline = "../tests/models/airline/model.txt"
    model_file_mtpl2 = "../tests/models/mtpl2/model.txt"

    run_benchmark(
        model_files=[model_file_NYC, model_file_airline, mtpl2_X],
        np_data=[NYC_X, airline_X, mtpl2_X],
        model_classes=[
            LLVMModel,
            LGBMModel,
            TreeliteModel,
            ONNXModel,
        ],
        threadcount=[1],
        batchsizes=[1, 2, 3, 5, 7, 10, 20, 30, 50, 70, 100, 200, 300],
        n_samples=20000,
    )

    run_benchmark(
        model_files=[model_file_mtpl2],
        np_data=[mtpl2_X],
        model_classes=[
            LLVMModel,
            LGBMModel,
            TreeliteModel,
            ONNXModel,
        ],
        threadcount=[4],
        batchsizes=[
            10000,
            100000,
            1000000,
        ],
        n_samples=100,
    )



import lightgbm
import numpy as np
import pytest
from hypothesis import given, settings
from hypothesis import strategies as st

import lleaves

MODEL_DIRS_NUMERICAL = [
    "tests/models/boston_housing/",
    "tests/models/NYC_taxi/",
    "tests/models/single_tree/",
    "tests/models/tiniest_single_tree/",
]


MODEL_DIRS_CATEGORICAL = [
    "tests/models/mixed_categorical/",
    "tests/models/pure_categorical/",
    "tests/models/airline/",
]
CAT_BITVEC_CATEGORICAL = [
    (True, True, True, False, False),
    (True, True, True),
    (True, True, True, True, False, False),
]


@pytest.fixture(scope="session", params=MODEL_DIRS_NUMERICAL)
def llvm_lgbm_model(request):
    path = request.param
    llvm = lleaves.Model(model_file=path + "model.txt")
    llvm.compile()
    return (
        llvm,
        lightgbm.Booster(model_file=path + "model.txt"),
    )


@pytest.fixture(
    scope="session", params=zip(MODEL_DIRS_CATEGORICAL, CAT_BITVEC_CATEGORICAL)
)
def llvm_lgbm_model_cat(request):
    path, bitvec = request.param

    llvm = lleaves.Model(model_file=path + "model.txt")
    llvm.compile()
    return (
        llvm,
        lightgbm.Booster(model_file=path + "model.txt"),
        bitvec,
    )


def test_attribute_similarity(llvm_lgbm_model):
    llvm_model, lightgbm_model = llvm_lgbm_model
    assert llvm_model.num_feature() == lightgbm_model.num_feature()


@settings(deadline=1000)
@given(data=st.data())
def test_forest_llvm_mode(data, llvm_lgbm_model):
    llvm_model, lightgbm_model = llvm_lgbm_model
    input_data = data.draw(
        st.lists(
            st.floats(allow_nan=True, allow_infinity=True),
            max_size=llvm_model.num_feature(),
            min_size=llvm_model.num_feature(),
        )
    )
    input_data = np.array([input_data])
    np.testing.assert_array_almost_equal(
        llvm_model.predict(input_data), lightgbm_model.predict(input_data)
    )


@settings(max_examples=10)
@given(data=st.data())
def test_batchmode(data, llvm_lgbm_model):
    llvm_model, lightgbm_model = llvm_lgbm_model
    input_data = []
    for i in range(10):
        input_data.append(
            data.draw(
                st.lists(
                    st.floats(allow_nan=True, allow_infinity=True),
                    max_size=llvm_model.num_feature(),
                    min_size=llvm_model.num_feature(),
                )
            )
        )
    input_data = np.array(input_data)
    np.testing.assert_almost_equal(
        llvm_model.predict(input_data), lightgbm_model.predict(input_data)
    )


@given(data=st.data())
@settings(deadline=None)  # the airline model takes a few seconds to compile
def test_forest_llvm_mode_cat(data, llvm_lgbm_model_cat):
    llvm_model, lgbm_model, cat_bitvec = llvm_lgbm_model_cat

    input_cats = data.draw(
        st.lists(
            st.integers(min_value=0, max_value=2 ** 31 - 2),
            min_size=sum(cat_bitvec),
            max_size=sum(cat_bitvec),
        )
    )
    input_floats = data.draw(
        st.lists(
            st.floats(allow_nan=True, allow_infinity=True),
            max_size=lgbm_model.num_feature() - sum(cat_bitvec),
            min_size=lgbm_model.num_feature() - sum(cat_bitvec),
        )
    )
    input_data = [
        input_cats.pop() if is_cat else input_floats.pop() for is_cat in cat_bitvec
    ]
    np.testing.assert_array_almost_equal(
        llvm_model.predict([input_data]), lgbm_model.predict([input_data]), decimal=15
    )



from os import path

from setuptools import find_packages, setup

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="lleaves",
    use_scm_version=True,
    setup_requires=["setuptools_scm"],
    packages=find_packages(exclude=["benchmarks"]),
    url="https://github.com/siboehm/lleaves",
    project_urls={"Documentation": "https://lleaves.readthedocs.io/en/latest/"},
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3",
    ],
    author="Simon Boehm",
    author_email="simon@siboehm.com",
    description="LLVM-based compiler for LightGBM models",
    long_description=long_description,
    long_description_content_type="text/markdown",
    python_requires=">=3.6",
    install_requires=["llvmlite>=0.36", "numpy"],
)



import lightgbm as lgb
import numpy as np
import pandas as pd

if __name__ == "__main__":
    # data can be downloaded from here: https://www.openml.org/d/1169
    # todo download data automatically
    df = pd.read_csv("data/airline_data.csv")
    y = df.pop("Delay")
    for c in ["Airline", "Flight", "AirportFrom", "AirportTo"]:
        df[c], _ = df[c].factorize()
    df.pop("Flight")
    df.to_csv("data/airline_data_factorized.csv", index=False)

    dataset = lgb.Dataset(
        df.astype(np.float32), label=y, categorical_feature=[0, 1, 2, 3]
    )
    model = lgb.train({"objective": "binary"}, dataset)
    model.save_model("../tests/models/airline/model.txt")



from lleaves.compiler.ast.scanner import scan_model_file


def test_parser():
    model_file = "tests/models/boston_housing/model.txt"
    result = scan_model_file(model_file)
    assert result["general_info"]["max_feature_idx"] == 12
    assert len(result["trees"]) == 100

    tree_3 = result["trees"][3]
    assert tree_3["num_leaves"] == 18
    assert tree_3["left_child"] == [
        1,
        3,
        -2,
        -1,
        9,
        8,
        7,
        10,
        15,
        16,
        -5,
        -9,
        -8,
        14,
        -14,
        -6,
        -3,
    ]

    tree_95 = result["trees"][95]
    assert tree_95["Tree"] == 95
    assert tree_95["num_leaves"] == 21

    model_file = "tests/models/tiniest_single_tree/model.txt"
    result = scan_model_file(model_file)
    assert len(result["trees"]) == 1
    tree_0 = result["trees"][0]
    assert tree_0["num_leaves"] == 4
